/* $Id: mcl_socket_lib.cpp,v 1.12 2004/12/15 17:30:18 roca Exp $ */
/*
 *  Copyright (c) 2004 INRIA - All rights reserved
 *  (main author: Vincent Roca - vincent.roca@inrialpes.fr)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#ifdef ALC
#include "../alc/mcl_includes.h"

#elif defined(NORM)

#include "../norm/mcl_includes.h"
#endif // RM_PROTOCOL


/******************************************************************************/
/*
 * Code derived from Richard Stevens's book:
 * "UNIX Network Programming, Volume 1, Second Edition: "Networking
 *  APIs: Sockets and XTI", Prentice Hall, 1998, ISBN 0-13-490012-X.
 * Available at URL: http://www.kohala.com/start/unpv12e.html
 */


/**
 * Set the multicast interface where to send or receiver multicast traffic.
 * => See header file for more informations.
 */
INT32
mcast_set_if	(MCL_SOCKET	sockfd,
		 INT32		af,
		 mcl_addr	*ifaddr,
		 const char	*ifname,
		 INT32		verbosity)
{
	INT32	err;

	if (verbosity >= 4) {
		PRINT_OUT((mcl_stdout,
			"mcast_set_if: af=%d, ifaddr=%s, ifname=%s\n", af,
			(ifaddr != NULL) ? ifaddr->get_addr_string() : "null",
			(ifname != NULL) ? ifname : "null"))
	}
	switch (af) {
	case AF_INET: {
		struct in_addr		inaddr;
		struct sockaddr_in	sin;	// used if ifaddr is provided
#ifndef WIN32	/* SIOCGIFADDR not defined in WinSock2? */
		struct ifreq		ifreq;	// used if ifname is provided

		if (ifname != NULL) {
			strncpy(ifreq.ifr_name, ifname, IF_NAMESIZE);
			if (ioctl(sockfd, SIOCGIFADDR, &ifreq) < 0) {
				/* i/f name not found */
				perror("mcast_set_if: ERROR, ioctl(SIOCGIFADDR)");
				PRINT_ERR((mcl_stderr,
					"mcast_set_if: ERROR, ioctl(SIOCGIFADDR) failed for ifname %s\n",
					ifname))
				errno = ENXIO;
				return(-1);
			}
			memcpy(&inaddr,
				&((struct sockaddr_in*)&ifreq.ifr_addr)->sin_addr,
				sizeof(struct in_addr));
		} else
#endif /* OS */
		if (ifaddr != NULL) {
			if (ifaddr->is_ipv4_addr() == false) {
				PRINT_ERR((mcl_stderr,
				"mcast_set_if: ERROR, ifaddr is IPv6 whereas the declared address family is IPv4!\n"))
				errno = EINVAL;
				return(-1);
			}
			ifaddr->get_addr_struct(&sin);
			memcpy(&inaddr, &((struct sockaddr_in*)&sin)->sin_addr,
				sizeof(struct in_addr));
		} else {
			/* Remove previous. Set interface to default */
			inaddr.s_addr = htonl(INADDR_ANY);
		}
		if ((err = setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_IF,
				(char*)&inaddr, sizeof(struct in_addr))) < 0) {
			int	e = errno;
			perror("mcast_set_if: ERROR, setsockopt(IP_MULTICAST_IF)");
			errno = e;
			return(-1);
		}
		return (err);
	}
#ifdef INET6
	case AF_INET6: {
		u_int	index;

#ifndef WIN32
		if (ifname == NULL) {
			/* must supply name */
			PRINT_ERR((mcl_stderr,
				"mcast_set_if: ERROR, ifname required\n"))
			errno = EINVAL;
			return(-1);
		}		
		if ((index = if_nametoindex(ifname)) == 0) {
			/* i/f name not found */
			perror("mcast_set_if: ERROR, if_nametoindex");
			errno = ENXIO;
			return(-1);
		}
#else
		index = 0;
#endif /* OS */
		if ((err = setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_IF,
				(char*)&index, sizeof(index))) < 0) {
			int	e = errno;
			perror("mcast_set_if: ERROR, setsockopt(IP_MULTICAST_IF)");
			errno = e;
			return(-1);
		}
		return (err);
	}
#endif
	default:
		PRINT_ERR((mcl_stderr,
		"mcast_set_if: ERROR, unsupported address family %d\n", af))
#ifndef WIN32
		errno = EPROTONOSUPPORT;
#endif  // !WIN32
		return(-1);
	}
}


/**
 * Set the TTL (with IPv4) or Hop Count (IPv6).
 * => See header file for more informations.
 */
INT32
mcast_set_ttl	(MCL_SOCKET	sockfd,
		 INT32		af,
		 INT32		val)
{
	INT32	err;

	switch (af) {
	case AF_INET: {
#ifdef WIN32
		UINT32	ttl;
#else
		/* solaris seems to care about the int/char type, not linux! */
		UINT8	ttl;
#endif

		ttl = val;
		if ((err = setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_TTL,
					(char*)&ttl, sizeof(ttl))) < 0) {
			perror("mcast_set_ttl: ERROR, setsockopt(TTL)");
		}
		break;
		}
#ifdef INET6
	case AF_INET6: {
		INT32	hops;
		hops = val;
		if ((err = setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
					(char*)&hops, sizeof(hops))) < 0) {
			perror("mcast_set_ttl: ERROR, setsockopt(HOPS)");
		}
		break;
		}
#endif
	default:
		PRINT_ERR((mcl_stderr,
		"mcast_set_ttl: ERROR, unsupported address family %d\n", af))
#ifndef WIN32
		errno = EPROTONOSUPPORT;
#endif  // !WIN32
		return(-1);
	}
	return err;
}


/**
 * Set/unset the loopback mode for multicast traffic.
 * => See header file for more informations.
 */
INT32
mcast_set_loop	(MCL_SOCKET	sockfd,
		 INT32		af,
		 INT32		onoff)
{
	switch (af) {
	case AF_INET: {
		UINT8	flag;
		flag = onoff;
		return(setsockopt(sockfd, IPPROTO_IP, IP_MULTICAST_LOOP,
				(char*)&flag, sizeof(flag)));
		}
#ifdef INET6
	case AF_INET6: {
		UINT32	flag;
		flag = onoff;
		return(setsockopt(sockfd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
				(char*)&flag, sizeof(flag)));
		}
#endif
	default:
		PRINT_ERR((mcl_stderr,
		"mcast_set_loop: ERROR, unsupported address family %d\n", af))
#ifndef WIN32
		errno = EPROTONOSUPPORT;
#endif  // !WIN32
		return(-1);
	}
}


/**
 * Join a multicast group.
 * => See header file for more informations.
 */
INT32
mcast_join (MCL_SOCKET		sockfd,
	    const sockaddr	*sa,
	    socklen_t		salen,
	    mcl_addr		*ifaddr,
	    const char		*ifname)
{
	switch (sa->sa_family) {
	case AF_INET: {
		struct ip_mreq		mreq;
		struct sockaddr_in	sin;	// used if ifaddr is provided
#ifndef WIN32	/* SIOCGIFADDR not defined in WinSock2? */
		struct ifreq		ifreq;
#endif /* OS */

		memcpy(&mreq.imr_multiaddr,
			   &((struct sockaddr_in *) sa)->sin_addr,
			   sizeof(struct in_addr));
			   
#ifndef WIN32
		if (ifname != NULL) {
			strncpy(ifreq.ifr_name, ifname, IFNAMSIZ);
			if (ioctl(sockfd, SIOCGIFADDR, &ifreq) < 0) {
				return(-1);
			}
			memcpy(&mreq.imr_interface,
				   &((struct sockaddr_in *) &ifreq.ifr_addr)->sin_addr,
				   sizeof(struct in_addr));
		} else
#endif /* OS */
		if (ifaddr != NULL) {
			if (ifaddr->is_ipv4_addr() == false) {
				PRINT_ERR((mcl_stderr,
				"mcast_join: ERROR, ifaddr is IPv6 whereas the declared address family is IPv4!\n"))
				errno = EINVAL;
				return(-1);
			}
			ifaddr->get_addr_struct(&sin);
			memcpy(&mreq.imr_interface,
				&((struct sockaddr_in*)&sin)->sin_addr,
				sizeof(struct in_addr));
		} else
			mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		return(setsockopt(sockfd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
				  (char*)&mreq, sizeof(mreq)));
		}
#ifdef INET6
	case AF_INET6: {
		struct ipv6_mreq	mreq6;

		memcpy(&mreq6.ipv6mr_multiaddr,
			   &((struct sockaddr_in6 *) sa)->sin6_addr,
			   sizeof(struct in6_addr));
#ifndef WIN32
		if (ifname != NULL) {
			if ( (mreq6.ipv6mr_interface = if_nametoindex(ifname)) == 0) {
				errno = ENXIO;	/* i/f name not found */
				return(-1);
			}
		} else {
			mreq6.ipv6mr_interface = 0;
		}
#else /* WIN32 */
		 mreq6.ipv6mr_interface = 0;
#endif /* OS */
	
#ifdef FREEBSD
		return(setsockopt(sockfd, IPPROTO_IPV6, IPV6_JOIN_GROUP,
				  (char*)&mreq6, sizeof(mreq6)));
#else
		return(setsockopt(sockfd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
				  (char*)&mreq6, sizeof(mreq6)));
#endif /* OS */
		}
#endif /* INET6 */
	default:
		PRINT_ERR((mcl_stderr,
			"mcast_join: ERROR, unsupported address family %d\n",
			sa->sa_family))
#ifndef WIN32
		errno = EPROTONOSUPPORT;
#endif  // !WIN32
		return(-1);
	}
}

